#include "global.h"
#include "sht11.h"
#include "delay.h"

/* 2 versions of glue code for crc check */
#ifdef SHT11_CRC_LOOKUP
#define  SHT11_CRC_CHECK(value) sht11_crc = PRG_RDB(&crcTable[sht11_crc^(value)]);
#else
#define  SHT11_CRC_CHECK(value) sht11_crc_check(value)
#endif

/* global variables of the sht11 sensor */
uint8_t sht11_crc;
uint8_t sht11_status_reg = 0;

#ifdef SHT11_CRC_LOOKUP
const uint8_t PROGMEM crcTable[] = {
  0, 49, 98, 83, 196, 245, 166, 151, 185, 136, 219, 234, 125, 76, 31, 46,
  67, 114, 33, 16, 135, 182, 229, 212, 250, 203, 152, 169, 62, 15, 92, 109,
  134, 183, 228, 213, 66, 115, 32, 17, 63, 14, 93, 108, 251, 202, 153, 168,
  197, 244, 167, 150, 1, 48, 99, 82, 124, 77, 30, 47, 184, 137, 218, 235,
  61, 12, 95, 110, 249, 200, 155, 170, 132, 181, 230, 215, 64, 113, 34, 19,
  126, 79, 28, 45, 186, 139, 216, 233, 199, 246, 165, 148, 3, 50, 97, 80,
  187, 138, 217, 232, 127, 78, 29, 44, 2, 51, 96, 81, 198, 247, 164, 149,
  248, 201, 154, 171, 60, 13, 94, 111, 65, 112, 35, 18, 133, 180, 231, 214,
  122, 75, 24, 41, 190, 143, 220, 237, 195, 242, 161, 144, 7, 54, 101, 84,
  57, 8, 91, 106, 253, 204, 159, 174, 128, 177, 226, 211, 68, 117, 38, 23,
  252, 205, 158, 175, 56, 9, 90, 107, 69, 116, 39, 22, 129, 176, 227, 210,
  191, 142, 221, 236, 123, 74, 25, 40, 6, 55, 100, 85, 194, 243, 160, 145,
  71, 118, 37, 20, 131, 178, 225, 208, 254, 207, 156, 173, 58, 11, 88, 105,
  4, 53, 102, 87, 192, 241, 162, 147, 189, 140, 223, 238, 121, 72, 27, 42,
  193, 240, 163, 146, 5, 52, 103, 86, 120, 73, 26, 43, 188, 141, 222, 239,
  130, 179, 224, 209, 70, 119, 36, 21, 59, 10, 89, 104, 255, 206, 157, 172 };
#endif

#ifndef SHT11_CRC_LOOKUP
void
sht11_crc_check(uint8_t value) {
  uint8_t i;
  for (i=8;i;i--) {
    if ((sht11_crc ^ value) & 0x80) {
      sht11_crc <<= 1;
      sht11_crc ^= 0x31;
    }
    else {
      sht11_crc <<= 1;
    }
    value <<=1;
  }
}
#endif

void
sht11_init( void ) {
  // SCK line as output but set to low first
  SHT11_SCK_PORT &=~(1<<SHT11_SCK_BIT);
  SHT11_SCK_DDR |= (1<<SHT11_SCK_BIT);

  // DATA to input. External pull up.
  // Set PORT to 0 => pull data line low by setting port as output
  SHT11_DATA_PORT &=~(1<<SHT11_DATA_BIT);
  SHT11_DATA_DDR |= (1<<SHT11_DATA_BIT);
}

void
sht11_delay() {
  uint8_t i= SHT11_DELAY_CYCLES;
  for (;i;i--)
    asm volatile("; loop\n\t"::);  /* required to prevent compiler optimisation */
}

void
sht11_reset() {
// communication reset: DATA-line=1 and at least 9 SCK cycles followed by transstart
//          _____________________________________________________         ________
// DATA: __|                                                     |_______|
//             _    _    _    _    _    _    _    _    _        ___     ___
// SCK : _____| |__| |__| |__| |__| |__| |__| |__| |__| |______|   |___|   |______
  uint8_t i;
  SHT11_DATA_HI();
  for (i=9;i;i--) {
    SHT11_SCK_HI();
    SHT11_SCK_LO();
  }
  sht11_transmission_start();
  sht11_sendbyte(SHT11_RESET);  // Soft reset
}

void
sht11_transmission_start( void ) {
// generates a transmission start
//       ____         ____
// DATA:     |_______|
//          ___     ___
// SCK : __|   |___|   |__
  SHT11_SCK_HI();
  SHT11_DATA_LO();
  SHT11_SCK_LO();
  SHT11_SCK_HI();
  SHT11_DATA_HI();
  SHT11_SCK_LO();
  // Reset crc calculation. Start value is the content of the status register.
  sht11_crc = sht11_mirrorbyte( sht11_status_reg & 0x0F );
}

uint8_t
sht11_mirrorbyte(uint8_t value) {
  uint8_t ret=0, i;
  for (i=0x80; i ; i>>=1) {
    if (value & 0x01)
      ret |= i;
    value >>= 1;
  }
  return ret;
}

uint8_t
sht11_readbyte( uint8_t send_ack ) {
  uint8_t mask;
  uint8_t value = 0;

  // SCK is low here !
  for (mask=0x80; mask; mask >>= 1 ) {
    SHT11_SCK_HI();  // SCK hi
    if ( SHT11_GET_BIT() != 0 )  // and read data
      value |= mask;
    SHT11_SCK_LO();   // SCK lo => sensor puts new data
  }

  /* send ACK if required */
  if ( send_ack ) {
      SHT11_DATA_LO();  // Get DATA line
  }
  SHT11_SCK_HI();    // give a clock pulse
  SHT11_SCK_LO();
  if ( send_ack ) {       // Release DATA line
    SHT11_DATA_HI();
  }

  return value;
}

uint8_t
sht11_sendbyte( uint8_t value ) {
  uint8_t mask;
  uint8_t ack;

  for (mask = 0x80; mask; mask>>=1) {
      SHT11_SCK_LO();

      if ( value & mask ) {  
	SHT11_DATA_HI(); 
      }
      else {
	SHT11_DATA_LO();
      }

      SHT11_SCK_HI();  // SCK hi => sensor reads data
    }
  SHT11_SCK_LO();

  // Release DATA line
  SHT11_DATA_HI();
  SHT11_SCK_HI();
  // Read acknowlege
  // ack = SHT11_GET_BIT() == 0 ? TRUE : FALSE;  // compiler doesn't like this
  ack = 0;
  if (!SHT11_GET_BIT())
    ack = 1;
  SHT11_SCK_LO();

  SHT11_CRC_CHECK(value);   // crc calculation
  return ack;
}

uint8_t
sht11_measure_start( sht11MeasureType type ) {
  // send a transmtion start and reset crc calculation
  sht11_transmission_start();
  // send command. Crc gets updated!
  return sht11_sendbyte( (uint8_t) type );
}

uint8_t
sht11_get_measure_value( uint16_t * value ) {
  uint8_t * chPtr = (uint8_t*) value;
  uint8_t checksum;
  uint8_t delay_count=62;  /* delay is 62 * 4ms */

  /* Wait for measurement to complete (DATA pin gets LOW) */
  /* Raise an error after we waited 250ms without success (210ms + 15%) */
  while ( SHT11_GET_BIT() ) {
    delay(1);
    delay_count--;
    if (delay_count == 0)
      return FALSE;
  }

  *(chPtr + 1) = sht11_readbyte( TRUE );  // read hi byte
  SHT11_CRC_CHECK(*(chPtr + 1));  // crc calculation
  *chPtr = sht11_readbyte( TRUE );    // read lo byte
  SHT11_CRC_CHECK(* chPtr);    // crc calculation

  checksum = sht11_readbyte( FALSE );   // crc calculation
  // compare it.
  return sht11_mirrorbyte( checksum ) == sht11_crc ? TRUE : FALSE;
}

uint8_t
sht11_status_write( uint8_t value ) {
  uint8_t ack;

  // send a transmission start and reset crc calculation
  sht11_transmission_start();

  // send command. Crc gets updated!
  ack = sht11_sendbyte( SHT11_STATUS_W );

  // send new status register value
  ack &= sht11_sendbyte( value );

  // if everything ok
  if ( ack )
    // Store new value. This is necessary since
    // contents of status register is initial value
    // of crc calculation.
    sht11_status_reg = value;

  return ack;
}

uint8_t
sht11_status_read( uint8_t *value ) {
  uint8_t checksum;

  // send a transmission start and reset crc calculation
  sht11_transmission_start();

  // send command. Crc gets updated!
  sht11_sendbyte( SHT11_STATUS_R );

  // read status register value. TRUE => send Ack
  *value = sht11_readbyte( TRUE );
  SHT11_CRC_CHECK(*value);
  //  sht11_crc = PRG_RDB(&crcTable[sht11_crc ^ *value]);

  // read checksum. FALSE => no Ack
  checksum = sht11_readbyte( FALSE );

  // if everything ok
  if ( sht11_mirrorbyte( checksum ) == sht11_crc ) {
      // Store new value. This is necessary since
      // contents of status register is initial value
      // of crc calculation.
      sht11_status_reg = *value;

      return TRUE;
    }
  else
    return FALSE;
}

uint16_t
sht11_humidity_scale(uint16_t value) {
  /* scaling factors as defined in the datasheet with factor 100 */
  /*-4.0 * 100*/
#define SHT11_C1 -400
  /*0.0405 * 100 * (1<<28)*/
#define SHT11_C2  1087163597
  /*-0.0000028 * 100 * (1<<30)*/
#define SHT11_C3 -300648
  /* humidity = SHT11_C1 + value * ( SHT11_C2 + SHT11_C3 * value); */
  /* I will calc (humidity * 100) here, the same scale that temperature has. */
  /* every step is calculated in the best possible scale */
  /* we calculate in long -> we have 31 bits + sign */
  /* value is 12 bit long */
  return (int32_t)
    (((((SHT11_C3 * value
	 >>2)               /* rescale 30 -> 28 */
	+ SHT11_C2)
       >> 11)               /* rescale 28 -> 17 */
      * value
      + ((int32_t)1<<16))   /* round: add (1<<16) */
     >> 17)                 /* rescale 17 -> 0 */
    + SHT11_C1;
}
