/* @file
 * @author Alex Raimondi
 * @date   8. 6. 2002
 *
 * @brief  Low level function to interface an SHT11.
 *
 * The SHT11 is a digital humidity & temperature sensmitter
 * from Sensirion (www.sensirion.com).
 */

/*
 * modified by Werner Hoch <werner.ho@gmx.de> for usage with avr-gcc
 * - added scaling and linearisation for humidity
 * - added crc calculation without lookup table
 * - added sht11_delay()
 * - added sht11_reset()
 * - implemented byte mirror in C instead of assembler
 */

/* Port definitions for SHT11 SCK pin */
#define SHT11_SCK_PORT	PORTD		// Port register
#define SHT11_SCK_DDR	DDRD		// Port direction register
#define SHT11_SCK_PIN	PIND		// Port input register
#define SHT11_SCK_BIT	PD5		// Bit in Port

/* Port definitions for SHT11 SCK pin */
#define SHT11_DATA_PORT	PORTD		// Port register
#define SHT11_DATA_DDR	DDRD            // Port direction register
#define SHT11_DATA_PIN	PIND		// Port input register
#define SHT11_DATA_BIT	PD4		// Bit in Port

/* define, whether use the bitwise crc algorithm (slower, less memory),
   or the lookuptable (faster, more memory (about 280Bytes)) */
//#define SHT11_CRC_LOOKUP

/* define how long to wait after a bit has been changed. This is
   necessary because of the external pullup and parasitic capacitances
   and depends on the circuit you use.
   Values between 0 and 255 are possible */
#define SHT11_DELAY_CYCLES 20

/* Macros to toggle port state of SCK line. */
#define SHT11_SCK_LO()	SHT11_SCK_PORT &= ~(1 << SHT11_SCK_BIT ) ; sht11_delay()
#define SHT11_SCK_HI()	SHT11_SCK_PORT |= (1 << SHT11_SCK_BIT ) ; sht11_delay()

/* Macros to toggle port state of DATA line.
 * Since DATA line has to be externaly pulled up by a resistor on initialisation
 * we have to write a 0 into port bit and toggle line by switching between
 * input and output. */
#define SHT11_DATA_LO()	SHT11_DATA_DDR |= (1 << SHT11_DATA_BIT) ; sht11_delay()
#define SHT11_DATA_HI()	SHT11_DATA_DDR &= ~(1 << SHT11_DATA_BIT) ; sht11_delay()

/* Macro to read DATA bit from port */
#define SHT11_GET_BIT() ((SHT11_DATA_PIN & (1<< SHT11_DATA_BIT)) > 0 ? 1 : 0)

/* Variable to hold crc value continously updated while communicating */
extern uint8_t sht11_crc;

/* Variable to hold current content of sht11 status register.
 * This value is needed as start value of crc calculation.
 * The initial value (after power on reset) is 0. */
extern uint8_t sht11_status_reg;

/* Definitions of all known sht11 commands */
#define SHT11_MEAS_T	0x03		// Start measuring of temperature.
#define SHT11_MEAS_RH	0x05		// Start measuring of humidity.
#define SHT11_STATUS_R	0x07		// Read status register.
#define SHT11_STATUS_W	0x06		// Write status register.
#define SHT11_RESET	0x1E		// Perform a sensor soft reset.

/* Enum to select between temperature and humidity measuring */
typedef enum _sht11MeasureType {
  sht11MeaT	= SHT11_MEAS_T,	// Temperature
  sht11MeaRh	= SHT11_MEAS_RH	// Humidity
} sht11MeasureType;

/* @brief	Start new transmission
 * Performs a transmission start sequence and resets crc
 * calculation.
 * @note	Do not use this function directly.  */
void sht11_transmission_start( void );

/* @brief	Read one byte from sensor
 * Reads one byte form sensor and sends or skips ACK
 * according to sendAck.
 * This function does not perform any crc calculation!
 * @note	Do not use this function directly.
 * @param	sendAck	if TRUE an ack will be sent to sensor
 * @return	the byte read form sensor */
uint8_t sht11_readbyte( uint8_t sendAck );

/* @brief	Send one byte to sensor
 * Sends a byte to the sensor and updates the crc calculation.
 * @note	Do not use this function directly.
 * @param	value	The value to be sent.
 * @return	TRUE if ACK form sensor received. FALSE otherwise. */
uint8_t sht11_sendbyte( uint8_t value );

/* @brief	Sets up sht11
 * Ports are set up to enable communication. */
void sht11_init( void );

/* @brief	Start measuring
 * Starts a temperature or humidity measurement according
 * to the value in type.
 * This function does not wait for the measurement to complete.
 * @param	type	The type of value that should be measured.
 * @return	TRUE if ACK form sensor received. FALSE otherwise. */
uint8_t sht11_measure_start( sht11MeasureType type );

/* @brief	Gets the result of measurement.
 * First this function checks for the sensor to be ready. 
 * It waits at most 250ms. If the sensor is not ready, it sends an error.
 * The value is read form sensor, the crc value is updated and compared with crc
 * value supplied by the sensor.
 * @param	value	Measured value is returned in here.
 * @return	TRUE if no crc error, FALSE otherwise. */
uint8_t sht11_get_measure_value( uint16_t * value );

/* @brief	Write status register
 * The supplied value is written into status register.
 * sht11StatusReg (needed for crc calculation) is updated, too.
 * @param	value	new status register value.
 * @return	TRUE if ACK form sensor received. FALSE otherwise. */
uint8_t sht11_status_write( uint8_t value );

/* @brief	Read status register
 * Reads out the sensor's status register. Crc is updated and compared
 * to the crc value supplied by the sensor.
 * sht11StatusReg (needed for crc calculation) is updated, too.
 * @param	value	Current status register value is returned in here.
 * @return	TRUE if no crc error, FALSE otherwise. */
uint8_t sht11_status_read( uint8_t *value );

/* reset the device */
void sht11_reset();

/* delay loop */
void sht11_delay();

/* mirrors one byte bit0 to bit7 and vice versa */
uint8_t sht11_mirrorbyte(uint8_t value);

/* scales and linearize the 12bit humiditiy value to humidity*100 */
/* the calculation error of the routine is -0.531 to 0.497 LSB */
uint16_t sht11_humidity_scale(uint16_t value);

/* crc calculation without using a lookup table */
#ifndef SHT11_CRC_LOOKUP
void sht11_crc_check(uint8_t value);
#endif
